import shutil
from pathlib import Path


def copy(src, dst):
    p_src = Path(src)
    p_dst = Path(dst)
    if not p_src.exists():
        raise FileNotFoundError(p_src.as_posix())
    p_dst.parent.mkdir(parents=True, exist_ok=True)
    if p_src.is_file():
        shutil.copy(src, dst)
    else:
        shutil.copytree(src, dst, dirs_exist_ok=True)




import traceback

import pkg_resources


def get_top_level(library_name):
    distribution = pkg_resources.get_distribution(library_name)
    return distribution.get_metadata_lines("top_level.txt")


def get_project_name(library_name):
    try:
        # 获取指定库的分发信息
        distribution = pkg_resources.get_distribution(library_name)
        return distribution.project_name
    except pkg_resources.DistributionNotFound:
        return library_name


def get_top_level_name(library_name):
    try:
        return list(get_top_level(library_name))[0]
    except FileNotFoundError:
        return get_project_name(library_name)
    except:
        traceback.print_exc()
        return library_name


# def get_dependencies(library_name):
#     try:
#         # 获取指定库的分发信息
#         distribution = pkg_resources.get_distribution(library_name)
#         depand_library_list = []
#         # 获取依赖库列表
#         dependencies = distribution.requires()
#         for depand in dependencies:
#             depand_library_list.append(get_top_level_name(depand.project_name))
#         return depand_library_list
#     except pkg_resources.DistributionNotFound:
#         # 如果找不到指定库，可以处理异常或返回错误消息
#         print(f"Library '{library_name}' not found.")
#         return []


def get_dependencies(library_name):
    try:
        # 获取指定库的分发信息
        distribution = pkg_resources.get_distribution(library_name)
        depand_library_list = []
        # 获取依赖库列表
        dependencies = distribution.requires()
        for depand in dependencies:
            # depand_library_list.append(get_top_level_name(depand.project_name))
            depand_library_list.append(depand.project_name)
        return list(set(depand_library_list))
    except pkg_resources.DistributionNotFound:
        # 如果找不到指定库，可以处理异常或返回错误消息
        print(f"Library '{library_name}' not found.1")
        return []


def get_version(library_name):
    try:
        # 获取指定库的分发信息
        distribution = pkg_resources.get_distribution(library_name)
        depand_library_list = []
        # 获取依赖库列表
        return distribution.version
    
    except pkg_resources.DistributionNotFound:
        traceback.print_exc()
        # 如果找不到指定库，可以处理异常或返回错误消息
        print(f"Library '{library_name}' not found.2")
        return '0.0.0'


if __name__ == "__main__":
    """
    """



from pathlib import Path

pack_root = Path().joinpath('pack_dir/Record')


class Dir(str):
    
    def __init__(self, sub_dir):
        self.sub_dir = sub_dir
    
    def str(self):
        return self.__str__()
    
    def __str__(self):
        return fr'{pack_root}\{self.sub_dir}'


Default_build_dir = Dir('Build')  # 默认打包路径
Default_module_dir = Dir('Module')  # 默认打包路径



import filecmp
import os
import shutil
import subprocess
import sys
import time
from pathlib import Path
from nuitka import Version
from NuFaster.nuitka_parser import np_parser
from NuFaster.setting import Default_build_dir
from NuFaster.tools import copy

pack_py = f'{__file__}/../pack_library.py'
Build_dir = Default_build_dir


def run_cmd(cmd):
    cmd = cmd.strip().replace('\t', ' '). \
        replace('\n', ' ').replace('  ', ' '). \
        replace('  ', ' ').replace('  ', ' ').replace('  ', ' ')
    print(cmd)
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
    a_time = time.time()
    while process.poll() is None:
        stdout_line = process.stdout.readline()
        
        if stdout_line:
            print(f"\rCMD Running: \033[91m{stdout_line.strip()}\033[0m"[:200], end=' ')
    print()
    b_time = time.time() - a_time
    print(f'CMD Run Complete: \033[91m{round(b_time, 2)}s\033[0m')


def get_base_dir():
    p_dir = Path(Build_dir).joinpath('Base')
    
    if not p_dir.exists():
        print('Packing Base .....')
        base_file = f'{pack_py}/../base.py'
        Path(base_file).parent.mkdir(parents=True, exist_ok=True)
        Path(p_dir).parent.mkdir(parents=True, exist_ok=True)
        with open(base_file, 'w') as f:
            f.write(f'print("Is Base")')
        dist_dir = to_pack(base_file, output_dir=Build_dir, standalone=True, show_progress=True)
        dist_dir.rename(p_dir)
        if Path(base_file).exists():
            try:
                Path(base_file).unlink(missing_ok=True)
            except:
                pass
    return p_dir


def get_dll_file(library_name, reload=False, del_py=True):
    from NuFaster.get_library_dll import Library
    library = Library(library_name)
    library_name = library.name
    
    Path(pack_py).parent.mkdir(parents=True, exist_ok=True)
    with open(pack_py, 'w') as f:
        f.write(f'import {library_name}\n')
        f.write(f'print("import {library_name}")')
    
    dll_dir = Path(Build_dir).joinpath(library.name, library.version, 'Dll')
    if dll_dir.exists() and not reload:
        return dll_dir
    
    # nofollow_list = []
    # if p_file.is_dir():
    #     if p_file.joinpath('tests').is_dir():
    #         nofollow_list.append(f"{module_name}.tests")
    #     for item in p_file.iterdir():
    #         if item.is_dir() and item.joinpath('tests').is_dir():
    #             nofollow_list.append(f"{module_name}.{item.stem}.tests")
    # nofollow_info = ','.join(nofollow_list)
    
    kwargs = {"--include-module"   : library_name,
              "--nofollow-imports" : True,
              '-nofollow-import-to': [*library.dependenciesName, f"{library_name}.tests.*", f"{library_name}.*.tests.*"]}
    
    dist_dir = to_pack(pack_py, output_dir=Build_dir, standalone=True, show_progress=True, **kwargs)
    if Path(pack_py).exists() and del_py:
        try:
            Path(pack_py).unlink(missing_ok=True)
        except:
            pass
    
    base_dir = get_base_dir()
    file_list = compare_folders(base_dir, dist_dir)
    for file in file_list[:]:
        if file.endswith('.exe'):
            file_list.remove(file)
        elif file in ['python3.dll']:
            file_list.remove(file)
    
    print(file_list)
    
    for file in file_list:
        p_file = Path(dist_dir).joinpath(file)
        p_dll_file = Path(dll_dir).joinpath(file)
        if not p_dll_file.exists():
            p_dll_file.parent.mkdir(parents=True, exist_ok=True)
            # print(p_file, p_dll_file)
            copy(p_file, p_dll_file)
    
    return dll_dir


def get_pyd_file(library_name):
    return to_pack_library_pyd(library_name)


def get_library_init_import(library_name):
    res_list = []
    __import__(library_name)
    modules = sys.modules
    for name, path in modules.items():
        if name.startswith(f'{library_name}.'):
            res_list.append(name)
    return res_list


def get_library_imports(library_name):
    from NuFaster.get_library_dll import Library
    library = Library(library_name)
    
    library_file = Path(library.library_file)
    if library_file.name == '__init__.py':
        library_file = library_file.parent
    
    else:
        pass


def to_pack(pack_file, output_dir=None, **kwargs):
    kwargs.get('mingw', False)
    kwargs.get('standalone', False)
    kwargs.get('show_progress', False)
    
    if output_dir:
        pack_res_dir = Path(output_dir).joinpath(Path(pack_file).stem + '.dist')
    else:
        pack_res_dir = Path().joinpath(Path(pack_file).stem + '.dist')
    kwargs['output_dir'] = output_dir
    usages = np_parser.parser_kwargs(kwargs)
    usages.append(pack_file)
    # pack_cmd = f"""
    #     nuitka
    #     --mingw
    #     --standalone
    #     --show-progress
    #     --output-dir={output_dir}
    #     {pack_file}
    # """
    # pack_cmd = pack_cmd.replace('\n', ' ')
    pack_cmd = ' '.join(usages)
    run_cmd(pack_cmd)
    return pack_res_dir


def to_pack_main(main_py, dependent_libs=None, output_dir=None, **kwargs):
    kwargs.get('mingw', False)
    kwargs.get('standalone', False)
    kwargs.get('show_progress', False)
    kwargs.get('windows_icon_from_ico', False)
    follow_import_to = kwargs.get('follow_import_to', [])
    include_module = kwargs.get('include_module', [])
    
    dependent_list = []
    dependent_names = []
    
    from NuFaster.get_library_dll import Library
    
    if dependent_libs:
        for lib in dependent_libs:
            dependent_list.append(Library(lib))
    for lib in dependent_list:
        print(f'加载库文件{lib}')
        if lib.name in follow_import_to or lib.name in include_module:
            continue
        dll = lib.dll_dir
        pyd = lib.pyd
        lib.check_dependencies()
        dependent_names.append(lib.name)
    kwargs['nofollow_import_to'] = dependent_names
    dist_dir = to_pack(main_py, output_dir=output_dir, **kwargs)
    for lib in dependent_list:
        dll = lib.dll_dir
        pyd = lib.pyd
        copy(pyd, dist_dir.joinpath(pyd.name))
        if not dll.exists():
            continue
        copy(dll, dist_dir)
    
    os.startfile(dist_dir)


def to_pack_library_pyd(library_name):
    from NuFaster.get_library_dll import Library
    library = Library(library_name)
    library_name = library.name
    library_file = Path(library.library_file)
    if library_file.name == '__init__.py':
        library_file = library_file.parent
    else:
        pass
    output_dir = Path(Build_dir).joinpath(str(library_name), library.version)
    return pack_module(library_file, library_name, output_dir)
    
    # pyd_name = f'{library_name}.cp{sys.winver.replace(".", "")}*.pyd'
    #
    # pack_cmd = f"""
    #             nuitka
    #             --mingw
    #             --module
    #             --show-progress
    #             --output-dir={output_dir}
    #             --include-module={library_name}
    #             --nofollow-import-to={library_name}.tests,{library_name}.*.tests
    #             {library_file.as_posix()}
    # """
    #
    # pack_cmd = pack_cmd.replace('\n', ' ')
    #
    # run_cmd(pack_cmd)
    # files = list(output_dir.glob(pyd_name))
    # if not files:
    #     raise Exception('打包失败')
    # return files[0]


def pack_module(file, module_name=None, output_dir=None, with_mingw=True, remove_build=False):
    if not module_name:
        module_name = Path(file).stem
    if not output_dir:
        output_dir = Path('NuitkaPackDir').joinpath(module_name)
    
    pyd_name = f'{module_name}.cp{sys.winver.replace(".", "")}*.pyd'
    p_file = Path(file)
    nofollow_list = []
    if p_file.is_dir():
        if p_file.joinpath('tests').is_dir():
            nofollow_list.append(f"{module_name}.tests")
        for item in p_file.iterdir():
            if item.is_dir() and item.joinpath('tests').is_dir():
                nofollow_list.append(f"{module_name}.{item.stem}.tests")
    nofollow_info = ','.join(nofollow_list)  # {f"--nofollow-import-to={nofollow_info}" if nofollow_info else ""}
    pack_cmd = f"""
                nuitka
                {"--mingw" if with_mingw else ""}
                {"--remove-output" if remove_build else ""}
                --module
                --show-progress
                --output-dir={output_dir}
                --include-module={module_name}
                --nofollow-import-to={module_name}.tests.*,{module_name}.*.tests.*
                {file}
    """
    
    pack_cmd = pack_cmd.replace('\n', ' ')
    
    run_cmd(pack_cmd)
    files = list(Path(output_dir).glob(pyd_name))
    if not files:
        raise Exception('打包失败')
    return files[0]


def compare_folders(folder1, folder2):
    dcmp = filecmp.dircmp(folder1, folder2)
    different_files = dcmp.right_only
    
    return different_files


if __name__ == '__main__':
    """
    Main run
    """
    # get_dll_file('opencv-python', reload=True, del_py=False)
    get_pyd_file('opencv-python')



from optparse import Option

from nuitka.OptionParsing import parser as _parser

print(_parser)


class Parser:
    
    @property
    def options(self):
        res_list = {}
        for option in _parser.option_list:
            option: Option
            long_opts = option._long_opts
            short_opts = option._short_opts
            for field in [*long_opts, *short_opts]:
                field = str(field).strip('-').replace('-', '_')
                if field in res_list:
                    print(field, 'is has')
                res_list[field] = option
        for group in _parser.option_groups:
            
            for option in group.option_list:
                option: Option
                long_opts = option._long_opts
                short_opts = option._short_opts
                for field in [*long_opts, *short_opts]:
                    field = str(field).strip('-').replace('-', '_')
                    if field in res_list:
                        print(field, 'is has')
                    res_list[field] = option
        return res_list
    
    def parser_kwargs(self, kwargs):
        usages = ['nuitka']
        options = self.options
        for key, val in kwargs.items():
            if val is None:
                continue
            key = str(key).strip('-').replace('-', '_')
            if key in options:
                option: Option = options.get(key)
                option_key = option._long_opts[0]
                if isinstance(val, bool):
                    if val:
                        usages.append(option_key)
                elif isinstance(val, str):
                    usages.append(f'{option_key}={val}')
                elif isinstance(val, (list, set, tuple)):
                    if not isinstance(option.default, list):
                        print('np_error: 关键字类型不为list')
                    val_str = ','.join([str(item) for item in val])
                    usages.append(f'{option_key}={val_str}')
        return usages


np_parser = Parser()

if __name__ == '__main__':
    """
    Main run
    """
    print(np_parser.options)



from setuptools import setup, find_packages

setup(
        name='NuFaster',  # 包的名称
        version='0.0.3',  # 版本号
        author='toca-Y',  # 作者名字
        author_email='',  # 作者邮箱
        description='借助nuitka更快的打包(分模块打包)',  # 包的简短描述
        long_description='分模块打包',  # 包的详细描述
        url='https://github.com/toca-Y/NuitkaPackV2',  # 包的主页或仓库链接
        packages=find_packages(),  # 寻找和包含所有的 Python 包
        install_requires=[
            'nuitka==1.5.8'
        ],  # 依赖项列表，如果有的话
        classifiers=[  # 包的分类
        ],
)



import shutil
import sys
from pathlib import Path

from NuFaster.get_library_depand import get_version, get_dependencies, get_top_level_name
from NuFaster.to_pack import get_pyd_file, get_dll_file
from NuFaster.setting import Default_module_dir
from NuFaster.tools import copy

py_version = sys.winver


class Library:
    
    def __init__(self, library_name, auto_generate=True):
        self.library_name = library_name
        self.name = get_top_level_name(library_name)
        self.auto_generate = auto_generate
        
        if self.version == '0.0.0':
            print(f'找不到{library_name}版本信息')
    
    def check_dependencies(self):
        """检查依赖库文件"""
        for lib in self.dependencies:
            print(f'检查依赖库: {lib}')
            lib_pyd = lib.pyd
            lib_dll = lib.dll_dir
            
            lib.check_dependencies()
            
            dep_pyd = self.dll_dir.joinpath(lib_pyd.name)
            
            if not dep_pyd.exists():
                print('复制依赖库的pyd文件')
                dep_pyd.parent.mkdir(parents=True, exist_ok=True)
                copy(lib_pyd, dep_pyd, )
                # shutil.copy(lib_pyd, dep_pyd, )
            
            # shutil.copytree(lib_dll, self.dll_dir, ignore_dangling_symlinks=True, dirs_exist_ok=True)
            
            for file in lib_dll.rglob('*'):
                relative = file.relative_to(lib_dll)
                
                dep_file = self.dll_dir.joinpath(relative)
                if dep_file.exists() and dep_file.stat().st_size == file.stat().st_size:
                    pass
                else:
                    # print(f'复制依赖库的dll文件:{relative}  {file} {file.exists()} -> {dep_file}')
                    print(f'复制依赖库的dll文件: {relative}')
                    dep_file.parent.mkdir(parents=True, exist_ok=True)
                    copy(file, dep_file)
                    # if file.is_file():
                    #     shutil.copy(file, dep_file)
                    # else:
                    #     shutil.copytree(file, dep_file, dirs_exist_ok=True)
    
    def reload_dll_file(self):
        shutil.rmtree(self.dll_dir, ignore_errors=True, )
        self.check_dependencies()
    
    @property
    def version(self):
        return get_version(self.library_name)
    
    @property
    def library_file(self):
        try:
            __import__(self.name)
        except ImportError as e:
            raise ImportError(f'{self.name}模块导入失败:{e}')
        module_info = sys.modules.get(self.name)
        return module_info.__file__
    
    @property
    def dependencies(self):
        res_list = []
        dependencies = get_dependencies(self.library_name)
        for lib in dependencies:
            res_list.append(Library(lib, auto_generate=self.auto_generate))
        return res_list
    
    @property
    def dependenciesName(self):
        res_list = []
        dependencies = get_dependencies(self.library_name)
        for lib in dependencies:
            res_list.append(get_top_level_name(lib))
        return res_list
    
    @property
    def dll_dir(self):
        dll_dir = Path(Default_module_dir.str()).joinpath(py_version, self.name, str(self.version), 'Dll')
        if dll_dir.exists():
            return dll_dir
        else:
            if not self.auto_generate:
                ok = input('当前不存在{self.name}的dll文件, 是否要生成0(否), 1(是):')
                if ok != '1':
                    return
            print(f'\n{self} 获取模块dll文件中')
            lib_dll_dir = get_dll_file(self.library_name)
            for file in Path(lib_dll_dir).rglob('*'):
                p_file = file
                relative = p_file.relative_to(lib_dll_dir)
                p_dll_file = Path(dll_dir).joinpath(relative)
                
                if not p_dll_file.exists():
                    p_dll_file.parent.mkdir(parents=True, exist_ok=True)
                    copy(p_file, p_dll_file)
                    # if p_file.is_file():
                    #     shutil.copy(p_file, p_dll_file)
                    # else:
                    #     shutil.copytree(p_file, p_dll_file)
        
        return dll_dir
    
    @property
    def pyd(self):
        pyd_name = f'{self.name}.pyd'
        pyd_file = Path(Default_module_dir).joinpath(py_version, self.name, str(self.version), pyd_name)
        if pyd_file.exists():
            return pyd_file
        else:
            if not self.auto_generate:
                ok = input('当前不存在{self.name}的pyd文件, 是否要生成0(否), 1(是):')
                if ok != '1':
                    return
            print(f'\n{self} 打包模块pyd文件中')
            pyd_file.parent.mkdir(parents=True, exist_ok=True)
            file = get_pyd_file(self.library_name)
            copy(file, pyd_file)
            # shutil.copy(file, pyd_file)
        return pyd_file
    
    def __str__(self):
        return f'< library: {self.name} >'


if __name__ == '__main__':
    """
    Main run
    """
    library = Library('pandas')
    print(library)
    # dll_ = library.dll_dir
    # print(dll_, dll_.exists())
    # pyd_ = library.pyd
    # print(pyd_, pyd_.exists())
    # library.check_dependencies()
